/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.components.modules.input;


import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.component.ComponentSelector;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.component.ComponentManager;
import org.apache.avalon.framework.component.Composable;
import org.apache.avalon.framework.logger.AbstractLoggable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.thread.ThreadSafe;

import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;

import org.apache.cocoon.components.modules.input.InputModule;

import org.apache.cocoon.matching.AbstractWildcardMatcher;

import java.util.ArrayList;
import java.util.Map;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * Constructs an array of values suitable for a JDBC collection type
 * from parameters obtained from another input module. Application is
 * not limited to JDBC collections but can be used wherever similar
 * named attributes shall be collected to an array of a given
 * type. Currently, long, int, and string are known, more to come.
 *
 * <p><b>Global and local configuration</b></p>
 * <table border="1">
 * <tr><td><code>input-module</code></td><td>Name of the input module used to obtain the value and its configuration</td></tr>
 * <tr><td><code>member</code></td>      <td>Collection member <table
 *                                           <tr><td>Attribute</td><td></td></tr>
 *                                           <tr><td>name</td><td>Parameter name, "*" may distinguish multiple collections</td></tr>
 *                                           <tr><td>type</td><td>JDBC type name of members</td></tr>
 *                                           </table>                       </td></tr>
 * </table>
 *
 * @author <a href="mailto:haul@informatik.tu-darmstadt.de">Christian Haul</a>
 * @version CVS $Id: CollectionMetaModule.java,v 1.1.2.1 2002/04/28 20:12:49 haul Exp $
 */
public class CollectionMetaModule extends AbstractLoggable
    implements InputModule, Configurable, Initializable, Composable, Disposable {

    /** The component manager instance */
    protected ComponentManager manager;

    protected Configuration memberConf = null;

    private String defaultInput = "request"; // default to request parameters
    private Configuration inputConf = null;  // will become an empty configuration object
                                             // during configure() so why bother here...
    String INPUT_MODULE_ROLE = InputModule.ROLE;
    String INPUT_MODULE_SELECTOR = INPUT_MODULE_ROLE+"Selector";

    private boolean initialized = false;
    private InputModule input = null;
    private ComponentSelector inputSelector = null;



    /**
     * Set the current <code>ComponentManager</code> instance used by this
     * <code>Composable</code>.
     */
    public void compose(ComponentManager manager) throws ComponentException {

        this.manager=manager;
    }


    
    public void configure(Configuration config) throws ConfigurationException {

        this.memberConf = config.getChild("member");
        this.inputConf = config.getChild("input-module");
        this.defaultInput = this.inputConf.getAttribute("name",this.defaultInput);
    }



    public void initialize() {

        try {
            // obtain input module
            this.inputSelector=(ComponentSelector) this.manager.lookup(INPUT_MODULE_SELECTOR); 
            if (this.defaultInput != null && 
                this.inputSelector != null && 
                this.inputSelector.hasComponent(this.defaultInput)
                ){
                this.input = (InputModule) this.inputSelector.select(this.defaultInput);
                if (!(this.input instanceof ThreadSafe && this.inputSelector instanceof ThreadSafe) ) {
                    this.inputSelector.release(this.input);
                    this.manager.release(this.inputSelector);
                    this.input = null;
                    this.inputSelector = null;
                }
                this.initialized = true;
            } else {
                if (getLogger().isErrorEnabled())
                    getLogger().error("A problem occurred setting up '" + this.defaultInput 
                                      + "': Selector is "+(this.inputSelector!=null?"not ":"")
                                      +"null, Component is "
                                      +(this.inputSelector!=null&&this.inputSelector.hasComponent(this.defaultInput)?"known":"unknown"));
            }
        } catch (Exception e) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("A problem occurred setting up '" + this.defaultInput + "': " + e.getMessage());
        }
    }



    public void dispose() {

        if (!this.initialized) 
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
        else 
            if (this.inputSelector != null) {
                if (this.input != null)
                    this.inputSelector.release(this.input);
                this.manager.release(this.inputSelector);
            }
    }


    

    public Object getAttribute( String name, Configuration modeConf, Request request ) 
        throws ConfigurationException {

        if (!this.initialized) {
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
            return null;
        }
        if (this.defaultInput == null) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("No input module given. FAILING");
            return null;
        }

        // obtain correct configuration objects
        // default vs dynamic
        Configuration mConf = this.memberConf;
        Configuration inputConfig = null;
        String inputName=null;
        if (modeConf!=null) {
            mConf       = modeConf.getChild("member");
            inputName   = modeConf.getChild("input-module").getAttribute("name",null);
            if (inputName != null) {
                inputConfig = modeConf.getChild("input-module");
            }
        }

        // read necessary parameters
        // name is used only if parameter name contains '*'
        // in that case it replaces '*' otherwise it is
        // ignored
        String jType = mConf.getAttribute("type","string");
        SortedSet matchset = new TreeSet();
        String pName = mConf.getAttribute("name");
        int index = pName.indexOf("*");
        if (index>-1) {
            String prefix = (index > 0 ? pName.substring(0,index) : null);
            String suffix = (index < (pName.length() -1) ? pName.substring(index+1,pName.length()) : null);
            pName = prefix+name+suffix;
        }

        getLogger().debug("jType "+jType);

        // done reading configuration
        // setup modules and read values
        Object result = null;
        try {
            Object[] values = null;
            if (this.input != null && inputName == null) {
                // input module is thread safe
                // thus we still have a reference to it
                // and
                // no other module is configured dynamically
                values = input.getAttributeValues(pName,this.inputConf,request);
            } else {
                // input was not thread safe
                // or
                // another module is configured dynamically
                // so acquire it again
                ComponentSelector iputSelector = null;
                InputModule iput = null;
                try {
                    // obtain input module
                    if (inputName == null) {
                        inputName   = this.defaultInput;
                        inputConfig = this.inputConf;
                    }

                    iputSelector=(ComponentSelector) this.manager.lookup(INPUT_MODULE_SELECTOR); 
                    if (this.defaultInput != null 
                        && iputSelector != null 
                        && iputSelector.hasComponent(inputName)) {
                        
                        iput = (InputModule) iputSelector.select(inputName);
                    }
                    if (iput != null) {
                        values = iput.getAttributeValues(pName, inputConfig, request);
                    }           
                } catch (Exception e) {
                    if (getLogger().isWarnEnabled()) 
                        getLogger().warn("A problem occurred acquiring a value from '" 
                                         + inputName + "' for '"+pName+"': " + e.getMessage());
                } finally {
                    // release components
                    if (iputSelector != null) {
                        if (iput != null)
                            iputSelector.release(iput);
                        this.manager.release(iputSelector);
                    }
                }

            }
            
            // done reading values
            // start converting values and assemble array

            if (values != null) {
                
                Object[] objects = new Object[values.length];
                    
                // FIXME: should put this into helper class
                for (int i = 0; i<values.length; i++) {
                    Object value = values[i];
                    if (jType.equalsIgnoreCase("string")) {
                        if (value instanceof String) {
                            objects[i] = value;
                        } else {
                            objects[i] = value.toString();
                        }
                    } else if (jType.equalsIgnoreCase("int")) {
                        if (value instanceof String) {
                            objects[i] = Integer.decode((String)value);
                        } else if (value instanceof Integer) {
                            objects[i] = value;
                        } else {
                            if (getLogger().isErrorEnabled())
                                getLogger().error("don't know how to cast");
                        }
                    } else if (jType.equalsIgnoreCase("long")) {
                        if (value instanceof String) {
                            objects[i] = Long.decode((String)value);
                        } else if (value instanceof Long) {
                            objects[i] = value;
                        } else {
                            if (getLogger().isErrorEnabled())
                                getLogger().error("don't know how to cast");
                        }
                    } else {
                        // other types need parsing & creation
                        if (getLogger().isErrorEnabled())
                            getLogger().error("don't know how to cast");
                    }
                    
                }
                return objects;
            }
        } catch (Exception e) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("A problem occurred acquiring a value from '" + inputName 
                                 + "' for '"+pName+"': " + e.getMessage());
        }
        return null;        
    }





    public Enumeration getAttributeNames( Configuration modeConf, Request request ) 
        throws ConfigurationException {

        if (!this.initialized) {
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
            return null;
        }
        if (this.defaultInput == null) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("No input module given. FAILING");
            return null;
        }

        // obtain correct configuration objects
        // default vs dynamic

        Configuration mConf = this.memberConf;
        Configuration inputConfig = null;
        String inputName=null;
        if (modeConf!=null) {
            mConf       = modeConf.getChild("member");
            inputName   = modeConf.getChild("input-module").getAttribute("name",null);
            if (inputName != null) {
                inputConfig = modeConf.getChild("input-module");
            }
        }

        // done reading configuration
        // setup modules and read attribute names
        Object result = null;
        try {
            Enumeration names = null;
            if (this.input != null && inputName == null) {
                // input module is thread safe
                // thus we still have a reference to it
                // and
                // no other module is configured dynamically
                names = input.getAttributeNames(this.inputConf,request);
            } else {
                // input was not thread safe
                // or
                // another module is configured dynamically
                // so acquire it again
                ComponentSelector iputSelector = null;
                InputModule iput = null;
                try {
                    // obtain input module
                    if (inputName == null) {
                        inputName   = this.defaultInput;
                        inputConfig = this.inputConf;
                    }

                    iputSelector=(ComponentSelector) this.manager.lookup(INPUT_MODULE_SELECTOR); 
                    if (this.defaultInput != null 
                        && iputSelector != null 
                        && iputSelector.hasComponent(inputName)) {
                        
                        iput = (InputModule) iputSelector.select(inputName);
                    }
                    if (iput != null) {
                        names = iput.getAttributeNames(inputConfig, request);
                    }           
                } catch (Exception e) {
                    if (getLogger().isWarnEnabled()) 
                        getLogger().warn("A problem occurred acquiring a names from '" + inputName + "': " + e.getMessage());
                } finally {
                    // release components
                    if (iputSelector != null) {
                        if (iput != null)
                            iputSelector.release(iput);
                        this.manager.release(iputSelector);
                    }
                }

            }

            // done reading names
            // find attribute names matching configuration
            // and return an enumeration

            if (names != null) {
                SortedSet matchset = new TreeSet();
                String pName = mConf.getAttribute("name");
                int index = pName.indexOf("*");

                if (index>-1) {
                    // parameter name contains '*'
                    // find all strings that match this '*'
                    // return them in an enumeration

                    String prefix = (index > 0 ? pName.substring(0,index) : null);
                    String suffix = (index < (pName.length() -1) ? pName.substring(index+1,pName.length()) : null);

                    while (names.hasMoreElements()) {
                        String name = (String)names.nextElement();
                        if (name.startsWith(prefix) && name.endsWith(suffix)) {
                            String wildcard = name.substring(prefix.length());
                            wildcard = wildcard.substring(0,wildcard.length()-suffix.length());
                            matchset.add(wildcard);
                        }
                    }
                } else {
                    // parameter name without wildcard
                    // check that name is among available names
                    // and return it in that case
                    boolean done=false;
                    while (!done && names.hasMoreElements()) {
                        String name = (String)names.nextElement();
                        if (name.equals(pName)) {
                            matchset.add(pName);
                            done = true;
                        }
                    }
                }
                return new EnumerationHelper(matchset.iterator());
            } else {
                return null;
            }

        } catch (Exception e) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("A problem occurred acquiring names from '" + inputName + "': " + e.getMessage());
        }
        return null;        
    }




    public Object[] getAttributeValues( String name, Configuration modeConf, Request request ) 
        throws ConfigurationException {

        Enumeration names = this.getAttributeNames( modeConf, request );
        ArrayList values = new ArrayList();
        while (names.hasMoreElements()) {
            values.add(this.getAttribute((String) names.nextElement(),modeConf,request));
        }

        return values.toArray();

    }

}

class EnumerationHelper implements Enumeration {
    Iterator iter = null;
    public EnumerationHelper( Iterator i ) { this.iter = i; }
    public boolean hasMoreElements() { return this.iter.hasNext(); }
    public Object nextElement() { return this.iter.next(); }
}
