/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
/*
  <mode name="CollectionMeta">
     <member name="foo_*" type="int"/>
  </mode>

 */
package org.apache.cocoon.components.modules.input;


import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.component.ComponentSelector;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.component.ComponentManager;
import org.apache.avalon.framework.component.Composable;
import org.apache.avalon.framework.logger.AbstractLoggable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.thread.ThreadSafe;

import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;

import org.apache.cocoon.components.modules.input.InputModule;

import org.apache.cocoon.matching.AbstractWildcardMatcher;

import java.net.URLEncoder;

import java.util.ArrayList;
import java.util.Map;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/** Meta module that obtains values from other module and returns
 * message digest of value. Very useful for storing and checking
 * passwords. Input module configured through nested element
 * "input-module", message digest algorithm, security provider, salt,
 * and URL encoded output configurable through attributes
 * "algorithm", "provider", "salt", "encode" of configuration root
 * element. Defaults are "sha", null, "salt", and "false".
 *
 * @author <a href="mailto:haul@informatik.tu-darmstadt.de">Christian Haul</a>
 * @version CVS $Id: DigestMetaModule.java,v 1.1.2.1 2002/04/28 20:12:49 haul Exp $
 */
public class DigestMetaModule extends AbstractLoggable
    implements InputModule, Configurable, Initializable, Composable, Disposable {

    /** The component manager instance */
    protected ComponentManager manager;

    private String defaultInput = "request"; // default to request parameters
    private Configuration inputConf = null;  // will become an empty configuration object
                                             // during configure() so why bother here...
    String INPUT_MODULE_ROLE = InputModule.ROLE;
    String INPUT_MODULE_SELECTOR = INPUT_MODULE_ROLE+"Selector";

    private boolean initialized = false;
    private InputModule input = null;
    private ComponentSelector inputSelector = null;

    private String defaultAlgorithm = "SHA";
    private String defaultProvider = null;
    private String defaultSalt = "salt";
    private boolean defaultEncode = false;


    /**
     * Set the current <code>ComponentManager</code> instance used by this
     * <code>Composable</code>.
     */
    public void compose(ComponentManager manager) throws ComponentException {

        this.manager=manager;
    }


    
    public void configure(Configuration config) throws ConfigurationException {

        this.inputConf = config.getChild("input-module");
        this.defaultAlgorithm = this.inputConf.getAttribute("algorithm",this.defaultAlgorithm);
        this.defaultProvider = this.inputConf.getAttribute("provider",this.defaultProvider);
        this.defaultSalt = this.inputConf.getAttribute("salt",this.defaultSalt);
        this.defaultEncode = this.inputConf.getAttributeAsBoolean("encode",this.defaultEncode);
    }



    public void initialize() {

        try {
            // obtain input module
            this.inputSelector=(ComponentSelector) this.manager.lookup(INPUT_MODULE_SELECTOR); 
            if (this.defaultInput != null && 
                this.inputSelector != null && 
                this.inputSelector.hasComponent(this.defaultInput)
                ){
                this.input = (InputModule) this.inputSelector.select(this.defaultInput);
                if (!(this.input instanceof ThreadSafe && this.inputSelector instanceof ThreadSafe) ) {
                    this.inputSelector.release(this.input);
                    this.manager.release(this.inputSelector);
                    this.input = null;
                    this.inputSelector = null;
                }
                this.initialized = true;
            } else {
                if (getLogger().isErrorEnabled())
                    getLogger().error("A problem occurred setting up '" + this.defaultInput 
                                      + "': Selector is "+(this.inputSelector!=null?"not ":"")
                                      +"null, Component is "
                                      +(this.inputSelector!=null&&this.inputSelector.hasComponent(this.defaultInput)?"known":"unknown"));
            }
        } catch (Exception e) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("A problem occurred setting up '" + this.defaultInput + "': " + e.getMessage());
        }
    }



    public void dispose() {

        if (!this.initialized) 
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
        else 
            if (this.inputSelector != null) {
                if (this.input != null)
                    this.inputSelector.release(this.input);
                this.manager.release(this.inputSelector);
            }
    }


    

    public Object getAttribute( String name, Configuration modeConf, Request request ) 
        throws ConfigurationException {

        if (!this.initialized) {
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
            return null;
        }
        if (this.defaultInput == null) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("No input module given. FAILING");
            return null;
        }

        // obtain correct configuration objects
        // default vs dynamic
        Configuration inputConfig = null;
        String inputName=null;
        if (modeConf!=null) {
            inputName   = modeConf.getChild("input-module").getAttribute("name",null);
            if (inputName != null) {
                inputConfig = modeConf.getChild("input-module");
            }
        }

        // read necessary parameters
        String algorithm = modeConf.getAttribute("algorithm",this.defaultAlgorithm);
        String provider  = modeConf.getAttribute("provider" ,this.defaultProvider );
        String salt  = modeConf.getAttribute("salt" ,this.defaultSalt );
        boolean encode = modeConf.getAttributeAsBoolean("encode" ,this.defaultEncode );

        // done reading configuration
        // setup modules and read values
        Object result = null;
        try {
            Object value = null;
            if (this.input != null && inputName == null) {
                // input module is thread safe
                // thus we still have a reference to it
                // and
                // no other module is configured dynamically
                value = input.getAttribute(name,this.inputConf,request);
            } else {
                // input was not thread safe
                // or
                // another module is configured dynamically
                // so acquire it again
                ComponentSelector iputSelector = null;
                InputModule iput = null;
                try {
                    // obtain input module
                    if (inputName == null) {
                        inputName   = this.defaultInput;
                        inputConfig = this.inputConf;
                    }

                    iputSelector=(ComponentSelector) this.manager.lookup(INPUT_MODULE_SELECTOR); 
                    if (this.defaultInput != null 
                        && iputSelector != null 
                        && iputSelector.hasComponent(inputName)) {
                        
                        iput = (InputModule) iputSelector.select(inputName);
                    }
                    if (iput != null) {
                        value = iput.getAttribute(name, inputConfig, request);
                    }           
                } catch (Exception e) {
                    if (getLogger().isWarnEnabled()) 
                        getLogger().warn("A problem occurred acquiring a value from '" 
                                         + inputName + "' for '"+name+"': " + e.getMessage());
                } finally {
                    // release components
                    if (iputSelector != null) {
                        if (iput != null)
                            iputSelector.release(iput);
                        this.manager.release(iputSelector);
                    }
                }

            }
            
            // done reading value
            // get message digest implementation and compute hash

            if (value != null) {

                try {
                    MessageDigest md = (provider==null ? MessageDigest.getInstance(algorithm) : 
                                        MessageDigest.getInstance(algorithm,provider));
                    
                    md.update((salt+(value instanceof String? (String)value : value.toString())).getBytes());
                    if (encode) {
                        return URLEncoder.encode(new String(md.digest()));
                    } else {
                        return md.digest();
                    }
                } catch (NoSuchAlgorithmException nsae) {
                    if (getLogger().isWarnEnabled()) 
                        getLogger().warn("A problem occurred acquiring digest algorithm '" + algorithm 
                                         + (provider==null?"":"' from '"+provider) +"': " + nsae.getMessage());
                }
            }
        } catch (Exception e) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("A problem occurred acquiring a value from '" + inputName 
                                 + "' for '"+name+"': " + e.getMessage());
        }
        return null;        
    }





    public Enumeration getAttributeNames( Configuration modeConf, Request request ) 
        throws ConfigurationException {

         if (!this.initialized) {
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
            return null;
        }
        if (this.defaultInput == null) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("No input module given. FAILING");
            return null;
        }

        // obtain correct configuration objects
        // default vs dynamic
        Configuration inputConfig = null;
        String inputName=null;
        if (modeConf!=null) {
            inputName   = modeConf.getChild("input-module").getAttribute("name",null);
            if (inputName != null) {
                inputConfig = modeConf.getChild("input-module");
            }
        }

        // read necessary parameters

        // done reading configuration
        // setup modules and read values
        Object result = null;
        try {
            Enumeration names = null;
            if (this.input != null && inputName == null) {
                // input module is thread safe
                // thus we still have a reference to it
                // and
                // no other module is configured dynamically
                names = input.getAttributeNames(this.inputConf,request);
            } else {
                // input was not thread safe
                // or
                // another module is configured dynamically
                // so acquire it again
                ComponentSelector iputSelector = null;
                InputModule iput = null;
                try {
                    // obtain input module
                    if (inputName == null) {
                        inputName   = this.defaultInput;
                        inputConfig = this.inputConf;
                    }

                    iputSelector=(ComponentSelector) this.manager.lookup(INPUT_MODULE_SELECTOR); 
                    if (this.defaultInput != null 
                        && iputSelector != null 
                        && iputSelector.hasComponent(inputName)) {
                        
                        iput = (InputModule) iputSelector.select(inputName);
                    }
                    if (iput != null) {
                        names = iput.getAttributeNames(inputConfig, request);
                    }           
                } catch (Exception e) {
                    if (getLogger().isWarnEnabled()) 
                        getLogger().warn("A problem occurred acquiring a names from '" 
                                         + inputName + "': " + e.getMessage());
                } finally {
                    // release components
                    if (iputSelector != null) {
                        if (iput != null)
                            iputSelector.release(iput);
                        this.manager.release(iputSelector);
                    }
                }

            }
            
            // done reading value

            return names;

        } catch (Exception e) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("A problem occurred acquiring names from '" + inputName 
                                 + "' : " + e.getMessage());
        }
        return null;        
   }




    public Object[] getAttributeValues( String name, Configuration modeConf, Request request ) 
        throws ConfigurationException {

        if (!this.initialized) {
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
            return null;
        }
        if (this.defaultInput == null) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("No input module given. FAILING");
            return null;
        }

        // obtain correct configuration objects
        // default vs dynamic
        Configuration inputConfig = null;
        String inputName=null;
        if (modeConf!=null) {
            inputName   = modeConf.getChild("input-module").getAttribute("name",null);
            if (inputName != null) {
                inputConfig = modeConf.getChild("input-module");
            }
        }

        // read necessary parameters
        String algorithm = modeConf.getAttribute("algorithm",this.defaultAlgorithm);
        String provider  = modeConf.getAttribute("provider" ,this.defaultProvider );
        String salt  = modeConf.getAttribute("salt" ,this.defaultSalt );
        boolean encode = modeConf.getAttributeAsBoolean("encode" ,this.defaultEncode );

        // done reading configuration
        // setup modules and read values
        Object[] result = null;
        try {
            Object[] values = null;
            if (this.input != null && inputName == null) {
                // input module is thread safe
                // thus we still have a reference to it
                // and
                // no other module is configured dynamically
                values = input.getAttributeValues(name,this.inputConf,request);
            } else {
                // input was not thread safe
                // or
                // another module is configured dynamically
                // so acquire it again
                ComponentSelector iputSelector = null;
                InputModule iput = null;
                try {
                    // obtain input module
                    if (inputName == null) {
                        inputName   = this.defaultInput;
                        inputConfig = this.inputConf;
                    }

                    iputSelector=(ComponentSelector) this.manager.lookup(INPUT_MODULE_SELECTOR); 
                    if (this.defaultInput != null 
                        && iputSelector != null 
                        && iputSelector.hasComponent(inputName)) {
                        
                        iput = (InputModule) iputSelector.select(inputName);
                    }
                    if (iput != null) {
                        values = iput.getAttributeValues(name, inputConfig, request);
                    }           
                } catch (Exception e) {
                    if (getLogger().isWarnEnabled()) 
                        getLogger().warn("A problem occurred acquiring a value from '" 
                                         + inputName + "' for '"+name+"': " + e.getMessage());
                } finally {
                    // release components
                    if (iputSelector != null) {
                        if (iput != null)
                            iputSelector.release(iput);
                        this.manager.release(iputSelector);
                    }
                }

            }
            
            // done reading value
            // get message digest implementation and compute hash

            if (values != null) {
                try {
                    MessageDigest md = (provider==null ? MessageDigest.getInstance(algorithm) : 
                                        MessageDigest.getInstance(algorithm,provider));
                    
                    result = new Object[values.length];
                    for (int i=0; i<values.length; i++) {
                        md.update((salt + (values[i] instanceof String? (String)values[i] : 
                                          values[i].toString())).getBytes());
                        if (encode) {
                            result[i] = URLEncoder.encode(new String(md.digest()));
                        } else {
                            result[i] = md.digest();
                        }
                    }
                } catch (NoSuchAlgorithmException nsae) {
                    if (getLogger().isWarnEnabled()) 
                        getLogger().warn("A problem occurred acquiring digest algorithm '" + algorithm 
                                         + (provider==null?"":"' from '"+provider) +"': " + nsae.getMessage());
                }
                return result;
            }
        } catch (Exception e) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("A problem occurred acquiring a value from '" + inputName 
                                 + "' for '"+name+"': " + e.getMessage());
        }
        return null;        
    }

}
